program OsmKostek;
{$APPTYPE CONSOLE}
uses
  SysUtils;

type
  TState = array[1..3,1..3] of Byte;
  TSteps = array[1..4] of Byte;

var Cnt : Integer;

function DoStep(X,Y,Step : Byte; St : TState) : TState;
begin
  Result:=St;
  case Step of
    1 : begin
          Result[X,Y]:=Result[X,Y-1];
          Result[X,Y-1]:=0;
        end;
    2 : begin
          Result[X,Y]:=Result[X,Y+1];
          Result[X,Y+1]:=0;
        end;
    3 : begin
          Result[X,Y]:=Result[X-1,Y];
          Result[X-1,Y]:=0;
        end;
    4 : begin
          Result[X,Y]:=Result[X+1,Y];
          Result[X+1,Y]:=0;
        end;
  end;
end;

function IsDone(St : TState) : Boolean;
begin
  Result:=(St[1,1]=1) and (St[1,2]=2) and (St[1,3]=3) and
          (St[2,1]=4) and (St[2,2]=5) and (St[2,3]=6) and
          (St[3,1]=7) and (St[3,2]=8) and (St[3,3]=0);
end;

procedure FindEmpty(St : TState; var X,Y : Integer);
var I,J : Integer;
begin
  for I:=1 to 3 do begin
    for J:=1 to 3 do begin
      if St[I,J]=0 then begin
        X:=I;
        Y:=J;
        exit;
      end;
    end;
  end;
end;

function ComputeManhatan(St : TState) : Integer;
var I,J,X,Y : Integer;
begin
  Result:=1;
  for I:=1 to 3 do begin
    for J:=1 to 3 do begin
      if St[I,J]=0 then begin
        X:=3;
        Y:=3;
      end else begin
        X:=((St[I,J]+2) div 3);
        Y:=((St[I,J]+2) mod 3)+1;
      end;
      Inc(Result,Abs(I-X)+Abs(J-Y));
    end;
  end;
end;

function ComputeDifference(St : TState) : Integer;
var I,J : Integer;
begin
  Result:=1;
  for I:=1 to 3 do begin
    for J:=1 to 3 do begin
      if not (St[I,J]=((I-1)*3+J)) then Inc(Result);
    end;
  end;
end;

function ComputeSteps(DorM : Boolean; X,Y : Integer; St : TState) : TSteps;
var StVal : TSteps;
    I,J : Integer;
    B : Byte;
begin
  for I:=1 to 4 do Result[I]:=I;
  if DorM then begin
    {Different}
    if (Y>1) then StVal[1]:=ComputeDifference(DoStep(X,Y,1,St)) else StVal[1]:=0;
    if (Y<3) then StVal[2]:=ComputeDifference(DoStep(X,Y,2,St)) else StVal[2]:=0;
    if (X>1) then StVal[3]:=ComputeDifference(DoStep(X,Y,3,St)) else StVal[3]:=0;
    if (X<3) then StVal[4]:=ComputeDifference(DoStep(X,Y,4,St)) else StVal[4]:=0;
  end else begin
    {Manhatan}
    if (Y>1) then StVal[1]:=ComputeManhatan(DoStep(X,Y,1,St)) else StVal[1]:=0;
    if (Y<3) then StVal[2]:=ComputeManhatan(DoStep(X,Y,2,St)) else StVal[2]:=0;
    if (X>1) then StVal[3]:=ComputeManhatan(DoStep(X,Y,3,St)) else StVal[3]:=0;
    if (X<3) then StVal[4]:=ComputeManhatan(DoStep(X,Y,4,St)) else StVal[4]:=0;
  end;
  {---}
  for I:=1 to 4 do begin
    if StVal[I]=0 then Result[I]:=0;
  end;
  for I:=2 to 4 do begin
    for J:=I to 4 do begin
      if StVal[J-1]>StVal[J] then begin
        {Switch}
        B:=Result[J-1];
        Result[J-1]:=Result[J];
        Result[J]:=B;
        {Switch}
        B:=StVal[J-1];
        StVal[J-1]:=StVal[J];
        StVal[J]:=B;
      end;
    end;
  end;
end;

procedure WriteState(St : TState);
var X,Y : Integer;
begin
  for X:=1 to 3 do begin
    for Y:=1 to 3 do begin
      if St[X,Y]=0 then Write(' ') else Write(St[X,Y]);
    end;
    WriteLn;
  end;
  WriteLn('---');
end;

function FindDown(DorM : Boolean; St : TState; Level : Integer) : Boolean;
var Steps : TSteps;
    I,X,Y : Integer;
begin
  Inc(Cnt);
  Result:=False;
  WriteState(St);
  if not IsDone(St) then begin
    if Level<20 then begin
      FindEmpty(St,X,Y);
      Steps:=ComputeSteps(DorM,X,Y,St);
      for I:=1 to 4 do begin
        if (Steps[I]<>0) then begin
          if FindDown(DorM,DoStep(X,Y,Steps[I],St),Level+1) then begin
            Result:=True;
            Exit;
          end;
        end;
      end;
    end;
  end else begin
    Result:=True;
  end;
end;

function GenerateRandom(Steps : Integer) : TState;
var I,X,Y,R : Integer;
begin
  R:=1;
  for X:=1 to 3 do begin
    for Y:=1 to 3 do begin
      Result[X,Y]:=R;
      Inc(R);
    end;
  end;
  Result[3,3]:=0;
  I:=0;
  X:=3;
  Y:=3;
  while I<Steps do begin
    R:=Random(4)+1;
    if ((R=1) and (Y>1)) or
       ((R=2) and (Y<3)) or
       ((R=3) and (X>1)) or
       ((R=4) and (X<3))
    then begin
      Result:=DoStep(X,Y,R,Result);
      case R of
        1 : Y:=Y-1;
        2 : Y:=Y+1;
        3 : X:=X-1;
        4 : X:=X+1;
      end;
      Inc(I);
    end;
  end;
end;

const R30 : TState = ((0,2,3),(1,8,5),(4,7,6));  // Dif(40)  Man(40)
      R50 : TState = ((5,4,0),(7,3,2),(8,1,6));  // Dif(X)
      R2h : TState = ((3,2,0),(6,5,1),(8,4,7));  // Dif(X)
      R1k : TState = ((4,8,7),(5,6,2),(0,3,1));  // Dif(X)
      Rdk : TState = ((0,2,4),(6,8,5),(7,1,3));  // Dif(X)

var State : TState;
    PocState : TState;
    CntM, CntD : Integer;
//    S : String;
begin
  WriteLn('Osm kostek :');
  {Random Generation}
{ --- >
Osm kostek :
Random 30
 23
185
476
---
Random 50
54
732
816
---
Random 200
32
651
847
---
Random 1000
487
562
 31
---
Random 10000
 24
685
713
---
 < --- }
{  WriteLn('Random 30');
  WriteState(GenerateRandom(30));
  WriteLn('Random 50');
  WriteState(GenerateRandom(50));
  WriteLn('Random 200');
  WriteState(GenerateRandom(200));
  WriteLn('Random 1000');
  WriteState(GenerateRandom(1000));
  WriteLn('Random 10000');
  WriteState(GenerateRandom(10000));
  WriteLn('--------'); }
  {One
  State[1,1]:=0;
  State[1,2]:=1;
  State[1,3]:=3;
  State[2,1]:=5;
  State[2,2]:=2;
  State[2,3]:=6;
  State[3,1]:=4;
  State[3,2]:=7;
  State[3,3]:=8;
  {Two
  State[1,1]:=0;
  State[1,2]:=1;
  State[1,3]:=2;
  State[2,1]:=4;
  State[2,2]:=5;
  State[2,3]:=3;
  State[3,1]:=7;
  State[3,2]:=8;
  State[3,3]:=6;
  {---}
  {  ReadLn(S);
  State[1,1]:=Ord(S[1])-48;
  State[1,2]:=Ord(S[2])-48;
  State[1,3]:=Ord(S[3])-48;
  ReadLn(S);
  State[2,1]:=Ord(S[1])-48;
  State[2,2]:=Ord(S[2])-48;
  State[2,3]:=Ord(S[3])-48;
  ReadLn(S);
  State[3,1]:=Ord(S[1])-48;
  State[3,2]:=Ord(S[2])-48;
  State[3,3]:=Ord(S[3])-48;}
  Randomize;
  State:=GenerateRandom(50);
  CntM:=-1;
  CntD:=-1;
  PocState:=State;
  if FindDown(False,State,0) then begin
    CntM:=Cnt;
    WriteLn('OK');
  end;
  if FindDown(True,State,0) then begin
    CntD:=Cnt;
    WriteLn('OK');
  end;
  WriteState(PocState);
  WriteLn('(',CntM,'|',CntD,')');
  ReadLn;
end.
